#include "jspfile.h"
#define WINGDI_FUDGE
#include <winsock2.h>
#include <fstream>
using std::ifstream;

string read(istream& in, int length) {
    static char buffer[256];
    in.read(buffer, length);
    buffer[length] = 0;
    return string(buffer, length);
}

void write(ostream& out, string str) {
    out.write(str.c_str(), str.length());
}

byte readByte(istream& in) {
    return (byte) in.get();
}

namespace pack {
    inline string _packShort(short i) { i = htons(i); return string((char*)(&i), 2); }
    inline short _unpackShort(const string &s, int i = 0) { return ntohs(*((short*)(s.substr(i, 2).c_str()))); }
    inline string _packLong(int32_t i) { i = htonl(i); return string((char*)(&i), 4); }
    inline short _unpackLong(const string &s, int i = 0) { return ntohl(*((long*)(s.substr(i, 4).c_str()))); }

    string packShort(short i) { string s = _packShort(i); return s.substr(1, 1) + s.substr(0, 1); }
    short unpackShort(const string &s, int i = 0) { return _unpackShort(s.substr(i+1,1)+s.substr(i,1), 0); }
    string packLong(long i) { string s = _packLong(i); return s.substr(3,1) + s.substr(2,1) + s.substr(1,1) + s.substr(0,1); }
    //long unpackLong(const string &s, int i = 0) { return _unpackLong(s.substr(i+3,1)+s.substr(i+2,1)+s.substr(i+1,1)+s.substr(i,1), 0); }
    long unpackLong(const string &s, int i = 0) { return *((long*)(s.substr(i, 4).c_str())); }

    //long unpackULong(const string &s) { return () }
}

string format(long i) {
    const char* chars = "0123456789ABCDEF";
    byte* b = (byte*)(&i);
    string r = "";
    r += chars[b[0]/16]; r += chars[b[0]%16];
    r += ' '; r += chars[b[1]/16]; r += chars[b[1]%16];
    r += ' '; r += chars[b[2]/16]; r += chars[b[2]%16];
    r += ' '; r += chars[b[3]/16]; r += chars[b[3]%16];
    return r;
}

bool jspfile::load(jspfile& file, string filename) {

    /*
    cerr << "begin pack testing" << endl;
    cerr << "    FF FF = " << pack::unpackShort(string("\xFF\xFF", 2)) << endl;
    cerr << "    00 00 = " << pack::unpackShort(string("\x00\x00", 2)) << endl;
    cerr << "    9E 1B 00 00 = " << pack::unpackLong(string("\x9E\x1B\x00\x00", 4)) << endl;
    cerr << "    C7 A9 00 00 = " << pack::unpackLong(string("\xC7\xA9\x00\x00", 4)) << endl;
    cerr << "    unpack ^ = " << format(pack::unpackLong(string("\xC7\xA9\x00\x00", 4))) << endl;
    cerr << "end pack testing" << endl;
    */

    file.imageInfo.clear();
    file.images.clear();
    file.imageCount = 0;

    ifstream in(filename.c_str(), ios_base::in | ios_base::binary);
    if (!in) return false;

    file.imageCount = pack::unpackShort(read(in, 2));

    for (short i = 0; i < file.imageCount; ++i) {
        jspfile::info info;

        info.width = pack::unpackShort(read(in, 2));
        info.height = pack::unpackShort(read(in, 2));
        info.originX = pack::unpackShort(read(in, 2));
        info.originY = pack::unpackShort(read(in, 2));
        info.byteSize = abs(pack::unpackLong(read(in, 4)));

        info.mystery[0] = pack::unpackShort(read(in, 2));
        info.mystery[1] = pack::unpackShort(read(in, 2));

        file.imageInfo.push_back(info);
    }

    for (short i = 0; i < file.imageCount; ++i) {
        jspfile::info info = file.imageInfo[i];

        Bitmap sprite(info.width, info.height);
        sprite.clear(0);

        int pos = 0;
        // cout << "sprite " << i << ": ";
        while (pos < info.width * info.height) {
            byte b = readByte(in);

            if (b > 128) {
                pos += b - 128;
                // cout << toString((int)(b - 128)) << " * transparent, ";
            } else {
                for (byte j = 0; j < b; ++j) {
                    sprite.putpixel(pos % info.width, pos / info.width, readByte(in));
                    ++pos;
                }
                // cout << toString((int) b) << " * data, ";
            }
        }
        // fcout << endl;

        file.images.push_back(sprite);
    }
    return true;
}

bool jspfile::save(jspfile& file, string filename) {
    // RLE

    byte** rle_data = new byte* [file.imageCount];
    long* rle_sizes = new long [file.imageCount];

    for (short i = 0; i < file.imageCount; ++i) {
        Bitmap sprite = file.images[i];
        byte* temp = new byte [sprite->w * sprite->h];
        for (int y = 0; y < sprite->h; ++y) {
            for (int x = 0; x < sprite->w; ++x) {
                temp[y * sprite->w + x] = (byte) sprite.getpixel(x, y);
            }
        }

        byte* data = new byte [sprite->w * sprite->h * 3];
        int n = 0, pos = 0;
        // cout << "sprite " << i << ": ";
        while (n < sprite->h * sprite->w) {
            byte c = temp[n];
            if (c == 0) {
                int len = 0;
                while (n+len <= sprite->w*sprite->h && (n % sprite->w) + len < sprite->w && len < 127 && temp[n+len] == 0) ++len;
                data[pos++] = 128 + len;
                // cout << len << " * transparent, ";
                n += len;
            } else {
                int len = 0;
                while (n+len <= sprite->w*sprite->h && (n % sprite->w) + len < sprite->w && len < 127 && temp[n+len] != 0) ++len;
                data[pos++] = len;
                for (int j = 0; j < len; ++j) data[pos++] = temp[n+j];
                // cout << len << " * data, ";
                n += len;
            }
        }
        // cout << endl;

        rle_data[i] = data;
        rle_sizes[i] = pos;
        delete[] temp;
    }

    // Header output

    ofstream out(filename.c_str(), ios_base::out | ios_base::binary);
    if (!out) return false;
    write(out, pack::packShort(file.imageCount));

    for (short i = 0; i < file.imageCount; ++i) {
        jspfile::info& info = file.imageInfo[i];
        info.byteSize = rle_sizes[i];

        write(out, pack::packShort(info.width));
        write(out, pack::packShort(info.height));
        write(out, pack::packShort(info.originX));
        write(out, pack::packShort(info.originY));
        write(out, pack::packLong(info.byteSize));
        write(out, pack::packShort(info.mystery[0]));
        write(out, pack::packShort(info.mystery[1]));
    }

    // Data output & cleanup

    for (short i = 0; i < file.imageCount; ++i) {
        for (long j = 0; j < rle_sizes[i]; ++j) {
            out.put(rle_data[i][j]);
        }

        // if (out.tellp() > 171809) cerr << "spr " << i << endl;

        delete[] rle_data[i];
    }

    delete[] rle_data;
    delete[] rle_sizes;
    return true;
}
