#include "stuff.h"
#include "jspfile.h"

Bitmap recolor(Bitmap input, int choices[8], int bright) {
    Bitmap result(input->w, input->h);
    result.draw(input, 0, 0);
    for (int x = 0; x < result->w; ++x) {
        for (int y = 0; y < result->h; ++y) {
            int color = result.getpixel(x, y);
            if (color == 0) continue;
            int family = color/32, value = color%32;
            family = choices[family]; value += bright;
            if (value < 0) value = 0;
            if (value < 1 && family == 0) value = 1;
            if (value > 31) value = 31;
            color = family * 32 + value;
            result.putpixel(x, y, color);
        }
    }
    return result;
}

void formatcolor(Bitmap buffer, int x, int y, int color) {
    const char* names[] = { "GRAY   ", "GREEN  ", "BROWN  ", "BLUE   ", "RED    ", "YELLOW ", "PINK   ", "AQUA   " };
    string name = names[color / 32];
    buffer.textout(font, name, x, y, color + 24, color + 8);
}

int recolortool(jspfile& file, int& sprite) {
    // static stuff
    const int bgsize = 40, crosssize = 3;

    Bitmap background(SCREEN_W, SCREEN_H);
    background.clear(PAL_GRAY + 16);
    for (int x = 0; x < SCREEN_W; x += bgsize) {
        for (int y = 0; y < SCREEN_H; y += bgsize) {
            if ((x/bgsize + y/bgsize) % 2 == 0) background.fillrect(x, y, x + bgsize - 1, y + bgsize - 1, PAL_GRAY + 24);
        }
    }
    background.fillrect(0, 0, SCREEN_W - 1, 3*bgsize - 1, PAL_GRAY + 31);
    background.line(0, 3*bgsize - 1, SCREEN_W - 1, 3*bgsize - 1, PAL_GRAY);

    background.textout(font, "Recolor Tool", 3, 3, PAL_GRAY);
    background.textout(font, "Ctrl-R to exit, Ctrl A-H to cycle options", 3, 14, PAL_GRAY);
    background.textout(font, " (A) Recolor  GRAY    to GRAY", 3, 25, PAL_GRAY);
    background.textout(font, " (B) Recolor  GREEN   to GREEN", 3, 36, PAL_GRAY);
    background.textout(font, " (C) Recolor  BROWN   to BROWN", 3, 47, PAL_GRAY);
    background.textout(font, " (D) Recolor  BLUE    to BLUE", 3, 58, PAL_GRAY);
    background.textout(font, " (E) Recolor  RED     to RED", 3, 69, PAL_GRAY);
    background.textout(font, " (F) Recolor  YELLOW  to YELLOW", 3, 80, PAL_GRAY);
    background.textout(font, " (G) Recolor  PINK    to PINK", 3, 91, PAL_GRAY);
    background.textout(font, " (H) Recolor  AQUA    to AQUA", 3, 102, PAL_GRAY);
    background.textout(font, "Ctrl-L/R to adjust brightness", 403, 14, PAL_GRAY);
    background.textout(font, "Left/Right to select sprite", 403, 69, PAL_GRAY);
    background.textout(font, "Ctrl-P to toggle preview:", 403, 80, PAL_GRAY);
    background.textout(font, "Ctrl-X to execute recolor/brightness", 403, 91, PAL_GRAY);
    for (int i = 0; i < 8; ++i) {
        formatcolor(background, 115, 25 + 11*i, 32*i);
    }

    // main loop

    Bitmap buffer(SCREEN_W, SCREEN_H);
    int choices[8] = { 0, 1, 2, 3, 4, 5, 6, 7 };
    int bright = 0;
    bool preview = true;

    int done = RET_NONE;
    while (done == RET_NONE) {
        if (close_button_pressed) done = RET_QUIT;

        if (keypressed()) {
            int k = readkey();
            char ch = k & 0xff;
            int scan = k >> 8;
            bool ctrl = key[KEY_LCONTROL] || key[KEY_RCONTROL];

            if (scan == KEY_ESC) {
                done = RET_OK;
            } else if (scan == KEY_LEFT) {
                if (ctrl) {
                    if (bright > -32) --bright;
                } else {
                    if (sprite > 0) --sprite;
                }
            } else if (scan == KEY_RIGHT) {
                if (ctrl) {
                    if (bright < 32) ++bright;
                } else {
                    if (sprite < file.imageCount - 1) ++sprite;
                }
            } else if (ctrl) {
                ch += 'a' - 1;
                if (ch >= 'a' && ch <= 'h') {
                    choices[ch - 'a'] = (choices[ch - 'a'] + 1) % 8;
                } else if (ch == 'p') {
                    preview = !preview;
                } else if (ch == 'x') {
                    file.images[sprite] = recolor(file.images[sprite], choices, bright);
                } else if (ch == 'r') {
                    done = RET_OK;
                }
            }
        }

        buffer.draw(background, 0, 0);
        for (int i = 0; i < 8; ++i) {
            formatcolor(buffer, 203, 25 + 11*i, 32*choices[i]);
        }

        jspfile::info i = file.imageInfo[sprite];
        buffer.textout(font, "Bright: " + toString(bright), 403, 25, PAL_GRAY);
        buffer.textout(font, "Sprite " + toString(sprite + 1) + " of " + toString(file.imageCount), 403, 36, PAL_GRAY);
        buffer.textout(font, "Dimensions: " + toString(i.width) + " x " + toString(i.height), 403, 47, PAL_GRAY);
        buffer.textout(font, "Origin: (" + toString(i.originX) + ", " + toString(i.originY) + ")", 403, 58, PAL_GRAY);
        buffer.textout(font, preview ? "ON" : "OFF", 611, 80, PAL_GRAY);
        if (preview) {
            buffer.draw_sprite(recolor(file.images[sprite], choices, bright), 400 - i.originX, 360 - i.originY);
        } else {
            buffer.draw_sprite(file.images[sprite], 400 - i.originX, 360 - i.originY);
        }
        buffer.line(400 - crosssize, 360, 400 + crosssize, 360, PAL_GRAY);
        buffer.line(400, 360 - crosssize, 400, 360 + crosssize, PAL_GRAY);
        buffer.draw_to_screen();
    }

    return done;
}
